# coding: utf-8
from __future__ import unicode_literals

from .common import InfoExtractor
from ..compat import (
    compat_str,
    compat_http_client,
    compat_HTTPError,
    compat_urllib_request,
    compat_urllib_error,
)
from ..utils import (
    base_url,
    error_to_compat_str,
    ExtractorError,
    sanitized_Request,
    std_headers,
    update_Request,
    update_url_query,
    urljoin,
    YoutubeDLRedirectHandler,
)

from StringIO import StringIO
import email

def encode_hdr(string, encoding='iso8859-1', errors='strict'):
    return string if isinstance(string, bytes) else string.encode(encoding, errors)

class UrlHandle(StringIO):
    _code = 0
    _url = ''
    _info = {}

    from types import MethodType

    def __init__(self, text, code, url, info):
        StringIO.__init__(self, text)
        self._info = email.message_from_string(encode_hdr(info, errors='backslashreplace'))
        # interface required by Py2 CookieJar.extract_cookies(); Py3 just uses info()
        self._info.getheaders = self.MethodType(lambda s, n: s.get_all(n) or [], self._info)
        self._code = code
        self._url = url

    def _read_only(self):
        forbidden = PermissionError()
        raise compat_urllib_error.URLError(forbidden)

    def truncate(self, size=float('inf')):
        self._read_only()

    def write(self, str):
        self._read_only()

    def writelines(self, sequence):
        self._read_only()

    @property
    def mode(self):
        return 'r'

    @property
    def headers(self):
        return self._info

    def geturl(self):
        return self._url

    def info(self):
        return self._info

    def getcode(self):
        return self._code


import subprocess
import re
import sys
import zlib
from itertools import takewhile

class FakeHTTP(InfoExtractor):

    def _request_webpage(self, url_or_request, video_id, note=None, errnote=None, fatal=True, data=None, headers={}, query={}, expected_status=None):
        if isinstance(url_or_request, compat_urllib_request.Request):
            url_or_request = update_Request(
                url_or_request, data=data, headers=headers, query=query)
        else:
            if query:
                url_or_request = update_url_query(url_or_request, query)
            url_or_request = sanitized_Request(url_or_request, data, headers)

        # type may not be an attribute before Py3
        def req_type(req):
           return req.type if not callable(getattr(req, 'get_type')) else (req.type or req.get_type())

        if req_type(url_or_request) not in ('http', 'https', ):
            return super(FakeHTTP, self)._request_webpage(url_or_request, video_id, note=None, errnote=errnote, fatal=fatal, expected_status=expected_status)

        if note is None:
            self.report_download_webpage(video_id)
        elif note is not False:
            if video_id is None:
                self.to_screen('%s' % (note,))
            else:
                self.to_screen('%s: %s' % (video_id, note))

        # Some sites check X-Forwarded-For HTTP header in order to figure out
        # the origin of the client behind proxy. This allows bypassing geo
        # restriction by faking this header's value to IP that belongs to some
        # geo unrestricted country. We will do so once we encounter any
        # geo restriction error.
        if self._x_forwarded_for_ip and not url_or_request.has_header('X-Forwarded-For'):
            url_or_request.add_header('X-Forwarded-For', self._x_forwarded_for_ip)

        for h, v in std_headers.items():
             #if h.capitalize() not in url_or_request.headers and h != 'Accept-Encoding':
             if not url_or_request.has_header(h.capitalize()) and h != 'Accept-Encoding':
                 url_or_request.add_header(h, v)

        exceptions = [compat_urllib_error.URLError, compat_http_client.HTTPException] #, socket.error]

        try:
            cmd = ['wget', '-4', '-U ""', '-q', '-S', '-O-']
            method = url_or_request.get_method()
            if method != 'GET':
                cmd.append('--method=%s' % method)
            self._downloader.cookiejar.add_cookie_header(url_or_request)
            for h_k, h_v in url_or_request.header_items():
                cmd.append('--header=%s: %s' % (h_k, h_v))
            #if encoding:
            #    cmd.append('--remote-encoding=%s ' % encoding)
            cmd.append('--connect-timeout=%s ' % self._downloader._socket_timeout)
            if data:
                post1 = (method in ('GET', 'POST'))
                cmd.append('--%s-data=%s' %
                           ('post' if method == 'GET' else 'body', data))
            else:
                post1 = (method == 'POST')
            # if POSTing, manually handle redirect to GET
            if post1:
                cmd.append('--max-redirect=0')
            url = url_or_request.get_full_url()
            cmd.append(url)
            debug_printtraffic = self._downloader.params.get('debug_printtraffic', False)
            if debug_printtraffic:
                self._downloader._write_string(' '.join(cmd))
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            stderr = stderr.decode('utf-8', 'replace')
            # print('-> %d' % p.returncode)
            msg = stderr.splitlines() or ['']
            code = 0
            reason = ''
            #charset = 'ascii'
            encoding = None
            header_i = -1
            in_headers = False
            for i in range(len(msg)):
                # print '%d: %s' % (i, msg[i])
                m = re.match(
                        r'''(?x)
                            \s*Location:\s+(?P<url>[\S]+?)($|\s+\[following\])|
                            \s+HTTP/[\d.]+\s+(?P<code>[1-5]\d\d)(?:\s+(?P<reason>.*))?$|
                            \s+Content-Type:\s+(?P<charset>.+)$|
                            \s+Content-Encoding:\s+(?P<encoding>\S+)|
                            (?P<other>\S.*)$
                        ''', msg[i])
                if m:
                    #print m.groups()
                    if m.group('url'):
                        new_url = urljoin(url, m.group('url'))
                        if new_url:
                            url = new_url
                    elif m.group('other'):
                        in_headers = False
                    else:
                        if m.group('code'):
                            code = int(m.group('code'))
                            reason = m.group('reason') or ''
                            if not in_headers:
                                header_i = i+1
                                in_headers = True
                        elif m.group('encoding'):
                            encoding = m.group('encoding').lower()
                        if debug_printtraffic:
                            self._downloader._write_string(msg[i].lstrip())
                else:
                    if debug_printtraffic:
                        self._downloader._write_string(msg[i].lstrip())
            #print 'ENCODING: %s' % encoding
            content = stdout
            if p.returncode != 0:
                if p.returncode == 8:
                    if not (code >= 300 and code < 310):
                        raise compat_HTTPError(url, code, reason, None, None)
                elif p.returncode == 6:
                    raise compat_HTTPError(url, 401, msg[-1], None, None)
                else:
                    raise compat_http_client.BadStatusLine(msg[-1])
            #print ' '.join( '%02x' % ord(x) for x in content[:12])
            if encoding in ('gzip', 'deflate'):
                # 47 = 32+15 = format:zlib,gzip + windowsize:32768
                content = zlib.decompress(content, 47)
            elif encoding:
                self.report_warning('Unsupported Content-Encoding: %s' % encoding)
            rsp = UrlHandle(
                      content, code, url,
                      '\n'.join(
                          [h.strip() for h in takewhile(
                              lambda x: x[0].isspace(), msg[header_i:])]))
            self._downloader.cookiejar.extract_cookies(rsp, url_or_request)
            if post1 and code >= 300 and code < 310:
                redir = YoutubeDLRedirectHandler()
                r_req = redir.redirect_request(url_or_request, None, code, reason, rsp.headers, url)
                if r_req:
                    return self._request_webpage(r_req, video_id, note=note, errnote=errnote, fatal=fatal, data=None, headers={}, query={}, expected_status=expected_status)
            if False and debug_printtraffic:
                if not encoding:
                    encoding = self._guess_encoding_from_content(rsp.headers.get('Content-Type', ''), content)
                try:
                    content = content.decode(encoding, 'replace')
                except LookupError:
                    content = content.decode('utf-8', 'replace')
                self._downloader._write_string(content)
            return rsp

        except tuple(exceptions) as err:
            if isinstance(err, compat_HTTPError):
                if self._InfoExtractor__can_accept_status_code(err, expected_status):
                    # Retain reference to error to prevent file object from
                    # being closed before it can be read. Works around the
                    # effects of <https://bugs.python.org/issue15002>
                    # introduced in Python 3.4.1.
                    err.fp._error = err
                    return err.fp

            if errnote is False:
                return False
            if errnote is None:
                errnote = 'Unable to download webpage'

            errmsg = '%s: %s' % (errnote, error_to_compat_str(err))
            if fatal:
                raise ExtractorError(errmsg, sys.exc_info()[2], cause=err)
            else:
                self._downloader.report_warning(errmsg)
                return False
